/** SALSA/World Wide Computer Project
 *
 * $Id: RMSPDaemon.java,v 1.9 2002/04/03 01:09:53 cvarela Exp $
 *
 * Reference Server - Remote Message Sending Protocol Daemon
 * In order to deliver a message to a remote actor, messengers
 * reach this daemon (that can also host actors) to get the  
 * reference of the target actor. The rmspd is in charge to forward
 * messengers to the target's stage (rmspd's host or remote applet).
 *
 * On entrance of an actor to a stage, the stage's internal registry 
 * is updated with a pair: {relativeUAL, localReference}.  Its 
 * references to local actors are updated to their current values in
 * this stage.  At this point, the actor is started locally.  If the
 * actor is an instance of Messenger, it is started with and "act()" 
 * message.
 * 
 * By Gregory Haik and Carlos Varela.  v0.1  June, 1999
 *
 * cvarela:  As of Salsa 0.3, the RMSPDaemon serves as a Universal 
 *           Actor Stage as well.  No forwarding is thus necessary.
 *           Applet support coming later...    1999/11/16
 */
package wwc.messaging;

import java.net.*;
import java.io.*;
import java.util.Hashtable;
import salsa.language.*;
import wwc.util.*;

/**
 * Contains the main method of the RMSP daemon.
 *
 * @version $I$, $G$
 * @author Gregory Haik, Carlos Varela
 */

public class RMSPDaemon {
    static int listeningPortNumber = 4040; 					// Default port number.
    static ServerSocket serverSocket = null;				// The server listens via this...
    static String protocolVersion = RMSProtocol.VERSION;	// This server implements the first version of UAN protocol.
    static String serverVersion = new String("rmspd/0.1");	// Location server version.
    static Hashtable registry;
    ObjectInputGate gate;
    int localStagePortNumber;
    String localHost;

    public RMSPDaemon() {
        registry = new Hashtable();

        // The ObjectInputGate does all the work of listening.
        try {
            localHost = InetAddress.getLocalHost().getHostName();
            gate = new ObjectInputGate(listeningPortNumber, "onReceive", this);
            Thread.currentThread().join();
        } catch (Exception e) {
//System.out.println("!!!1");
            System.out.println(e);
        }
    }

    public void onReceive(Object o) {
//System.out.println("Received:"+o);
//System.out.print("Received: ");
        // If it's a message to be passed on...
	    if (o instanceof Message) {  // just serializing data
//System.out.println("A");
//System.out.println("Message "+o+" received.");
	        Message m = (Message) o;
	        m.updateReferences();
	        m.getTarget().send(m);
	    }

        // If it's an instance of an Actor that's trying to migrate to here...
    	if (o instanceof UniversalActor) {
//System.out.println("B");
//System.out.println("Actor "+o+" received.");
	        UniversalActor incomingActor = (UniversalActor) o;
	        String relativeUAL = incomingActor.getUAL().getIdentifier();
//if (registry.containsKey(relativeUAL))
//System.out.println("Warning:  relativeUAL "+relativeUAL+" already in RMSPD registry... overwriting");
            registry.put(relativeUAL, incomingActor);
            if (o instanceof Messenger)
    	        ((Messenger) o).updateReferences();
// cvarela: we need to do the same for incomingActor's actor references!!!
            RunTime.receivedUniversalActor();
            incomingActor.start();
        }

        // If it's an instance of a Messenger
    	if (o instanceof Messenger) {
//System.out.println("C");
	        Messenger m = (Messenger) o;
	        // SALSA:  m<-act();
	        {
		        Actor[] _targets = {m};
		        String[] _methodNames = {"act"};
		        Object[][] _arguments = {{}};
		        int [] _tokens = {-1};
		        try {
		            m.send
			            (Message.createMessage
			                (null,
                            _targets,
			                _methodNames,
                            _arguments,
                            _tokens));
        		} catch (NoSuchMessageException _nsme) {
//System.out.println("!!!2");
		            _nsme.printStackTrace();
		        } catch (MessageCreationException _mce) {
//System.out.println("!!!3");
					System.err.println("SALSA Internal Error:"+
						_mce.getMessage());
                }
            }
// 	        this.forwardMessenger(m); 
		}

// 	    if (className.equals("wwc.messaging.IncomingActorNotification"))
//      {
// 	        IncomingActorNotification ian = (IncomingActorNotification) o;
//   	    RegisteryEntry encapRegEntry = ian.getRegisteryEntry(); 
//   	    if (registery.containsKey(ian.getRelativeUAL()))
// 	    	    ian.complain("The relative UAL :\n "
// 		    		 + ian.getRelativeUAL()
// 			    	 + "\nis already used.");
//   	    else {
// 	    	registery.put(ian.getRelativeUAL(),
// 		    	      ian.getRegisteryEntry());
//   		String str = ian.getRegisteryEntry().toString();
// 	        }
// 	    }
// 	    else if (className.equals("wwc.messaging.EmigratingActorNotification")) {
// 	        EmigratingActorNotification ean = (EmigratingActorNotification) o;
// 	        registery.remove(ean.getRelativeUAL());
// 	    }
// 	    }
//        else if (className.equals("wwc.messaging.UnbindRequest")) {
//        UnbindRequest ur = (UnbindRequest) o;
//        registery.remove(ur.getRelativeUAL());
//   	}
// 	    System.out.println(registery.toString());
    }

	public static void main(String[] args) {
		// Handle arguments...
		if (args.length > 0) 
			if (args.length == 1) 
				if ((args[0].equals("-h")) || (args[0].equals("-?")))
					RMSPDaemon.helpMessage();
				else if (args[0].equals("-v"))
					RMSPDaemon.printServerVersion();
				else
					RMSPDaemon.helpMessage("illegal option");
			else if  (args.length != 2)
				RMSPDaemon.helpMessage("too many arguments");
			else if (!(args[0].equals("-p")))
				RMSPDaemon.helpMessage("wrong options");
			else
				try { 
					listeningPortNumber = Integer.parseInt(args[1]);
				} catch (Exception e) {
//System.out.println("!!!4");
					RMSPDaemon.helpMessage("invalid port number : " + args[1]);
				}

		System.out.println("rmspd started at "
			+(new java.sql.Time(System.currentTimeMillis())).toString()
			+"...");
		System.out.println("listening on port "+ new Integer(listeningPortNumber).toString());
		RMSPDaemon myDeamon = new RMSPDaemon();
	}

	public static int getListeningPortNumber() {
		return listeningPortNumber;
	}

	public static UniversalActor getActorUAL(String relativeUAL) {
		return (UniversalActor) registry.get(relativeUAL);
	}

//     private void forwardMessenger(Messenger m) {
// 	String targetRelativeUAL = m.getTargetUAL().getIdentifier();
// 	if (! registery.containsKey(targetRelativeUAL)) 
// 	    // cvarela: search for NEW location from target's UAN.
// 	    m.backToSender();
// 	else {
// 	    RegisteryEntry entry = (RegisteryEntry) registery.get(targetRelativeUAL);
// 	    m.setHashCode(entry.getActorHashCode());
// 	    if (entry.getAppletHostName() == null)
// 		this.exportMessenger(m);
// 	    else 
// 		this.exportMessenger(m,
// 				     entry.getAppletHostName(),
// 				     entry.getAppletPortNumber());
// 	}
//     }
    
//     private void exportMessenger(Messenger m) {
// 	this.exportMessenger(m, localHost, localStagePortNumber);
//     }

// /**
//  * Unsafe ! messenger might not be exported properly !!
//  *
//  */
//     private void exportMessenger(Messenger m,
// 				 String hostname,
// 				 int portNumber) {
// 	try {
// 	    Socket socket = new Socket(hostname, portNumber);
// 	    ObjectOutputStream out = new ObjectOutputStream(
// 				 new BufferedOutputStream(socket.getOutputStream()));
// 	    out.writeObject(m);
// 	    out.close();
// 	    socket.close();
// 	} catch (UnknownHostException e) {
// 	} catch (java.io.IOException e) {}
//     }
        
	private static void printServerVersion() {
		System.out.println("Remote Message Sending Protocol Daemon.");
		System.out.println(serverVersion);
		System.exit(0);
	}
    
	private static void helpMessage(String s) {
		System.out.println(s);
		RMSPDaemon.printHelpMessage();
		System.exit(1);
	}

	private static void helpMessage() {
		RMSPDaemon.printHelpMessage();
		System.exit(0);
	}

	private static void printHelpMessage() {
		System.out.println("usage :");
		System.out.println("  java ...RMSPDaemon");
		System.out.println("  java ...RMSPDaemon -h");
		System.out.println("  java ...RMSPDaemon -v");
		System.out.println("  java ...RMSPDaemon -p portNumber");
		System.out.println("options :");
		System.out.println("  -h : print this message");
		System.out.println("  -v : print version");
		System.out.println("  -p portNumber : set the listening port to portNumber");
		System.out.println("                  default port number is 3030");
	}   
}
